/******************************************************
STM8S003K3 16MHz, IAR1.4+STM8S_StdPeriph_Lib_V2.1.0
project: psu1
******************************************************/
#include "stm8s.h"
#include "iostm003.c"
#include "bitset.c"
#include <stdbool.h>
#include <stdlib.h>
#include <stm8s_itc.h>

#define A      PD_ODR_ODR5
#define B      PD_ODR_ODR4
#define C      PD_ODR_ODR3
#define D      PD_ODR_ODR2
#define E      PD_ODR_ODR1
#define F      PC_ODR_ODR6
#define G      PC_ODR_ODR5
#define DP     PC_ODR_ODR4
#define N0     PF_ODR_ODR4
#define N1     PA_ODR_ODR3
#define N2     PD_ODR_ODR7
#define N3     PD_ODR_ODR6

#define key0   PB_IDR_IDR5
#define key1   PB_IDR_IDR4
#define key2   PB_IDR_IDR7

#define mosfet PE_ODR_ODR5

#define RV_upr 2000 //Ohms
#define RV_lwr 510  //Ohms
#define V_max_adc ((500UL*(RV_upr+RV_lwr))/RV_lwr) //=U_max*100, 24.6V*100
#define I_max_adc 1000  //0x1ff adc points

enum btn_states {prs, rls, hold} btn0,btn1,btn2;
enum dmodes {start,regU,regI,ovl,scp,sdn} dev_st=start;

#define timnmbr 5
u8 tkey[timnmbr],tmr[timnmbr];
u16 tim[timnmbr];

u16 adc1val, adc0val;
u16 Udef, Idef, Uadc, Iadc; 
s16 U_tmp, I_tmp, def_tmp;;

bool i_fg;

u8 dbuf[4]={4,3,2,1};
u8 dpp=0;

#define msize 10
__no_init u16 eeIval [msize] @ 0x4000; 
__no_init u16 eeUval [msize] @ 0x4000+sizeof(u16)*10;

inline static void gpio_cfg() 
{
          PB_DDR=PC_DDR=PD_DDR=PE_DDR=PF_DDR=0xff; //conf as output
	PA_CR1=PB_CR1=PC_CR1=PD_CR1=PE_CR1=PF_CR1=0xff; //push-pull|pullup
  	//PA_CR2=PB_CR2=PC_CR2=PD_CR2=PE_CR2=PF_CR2=0;  //speed 2MHz|int disabled

	PA_ODR = 1<<2|1<<1; PA_DDR = 1<<3;
	PB_CR1 = binary(11110000); PB_ODR = binary(11000000); PB_DDR = 0;
   PC_ODR = 1<<7;
   PD_DDR_DDR0 = 0;  //BKIN

}
inline static void clk_cfg()
{
  	CLK_DeInit();
	CLK->CKDIVR = 0;     //16MHz
   CLK->PCKENR1 = CLK_PCKENR1_TIM1|CLK_PCKENR1_TIM2;
}
inline static void t1_cfg()
{
  {
    #define toppwm 0xff/2
    
    TIM1->ARRH = 0;
    TIM1->ARRL = toppwm;

    TIM1->CCR1H=0;
    TIM1->CCR1L=0;
    TIM1->CCR2H=0;
    TIM1->CCR2L=TIM1->ARRL;
  }

    TIM1->CR1 = TIM1_COUNTERMODE_CENTERALIGNED1|TIM1_CR1_ARPE|TIM1_CR1_URS;
    TIM1->CR2 = TIM1_CR2_CCPC;
    TIM1->RCR = 1;

    TIM1->CCMR1|= TIM1_OCMODE_PWM1|TIM1_CCMR_OCxPE;
    TIM1->CCMR2|= TIM1_OCMODE_PWM2|TIM1_CCMR_OCxPE;
    TIM1->CCER1|= TIM1_CCER1_CC1E;
    TIM1->CCER1|= TIM1_CCER1_CC2E;
    TIM1->EGR  |= TIM1_EGR_COMG;  //update CCER reg with COM event
    TIM1->IER   = TIM1_IER_UIE;   //update interrupt enable

    TIM1->BKR  = TIM1_BKR_MOE|TIM1_BKR_OSSR|TIM1_BKR_OSSI|TIM1_BKR_BKE;
    TIM1->OISR = 0;
    TIM1->CR1 |= TIM1_CR1_CEN;
    TIM1->BKR |= TIM1_LOCKLEVEL_3;
}
static inline void t2_cfg()
{
  TIM2->PSCR  = 1;
  TIM2->ARRH  = 6;
  TIM2->ARRL  = 0x40; //5000Hz
  TIM2->CR1   = TIM2_CR1_ARPE|TIM2_CR1_URS;
  TIM2->IER   = TIM2_IER_UIE;

  TIM2->CR1 |= TIM2_CR1_CEN;
}
inline static void adc_cfg()
{  
	ADC1->CR1 = ADC1_PRESSEL_FCPU_D8;
	ADC1->CR2 = ADC1_ALIGN_RIGHT|ADC1_CR2_SCAN; 
   ADC1->CR3 = ADC1_CR3_DBUF;
   ADC1->CSR = 1; //channel sel
   ADC1->TDRL = binary(00000011); //smtt trg on adc pins is off
	ADC1->CR1 |= ADC1_CR1_ADON;
  
}
inline static void iwdg_cfg()
{
IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
IWDG_SetPrescaler(IWDG_Prescaler_4);
IWDG_SetReload(0xff); 
IWDG_Enable();
IWDG_ReloadCounter();
}

void convert_data (s16 d)
{
u8 k;
 if(d>=0)
 {
  for (k=0;k<4;k++)
  {
   dbuf[k]=d%10;
   d/=10;
  }
 } 
 else
 {
  d=abs(d);
  for (k=0;k<4;k++)
  {
   dbuf[k]=d%10;
   d/=10;
  }
  dbuf[3]='-'; 
 }
}
void display_data ()
{
  static u8 pc;

  (pc)?(pc--):(pc = 3);
  N0=N1=N2=N3=0; DP=1;

  switch (dbuf[pc])
 {
  case 0: A=0; B=0; C=0; D=0; E=0; F=0; G=1; break;
  case 1: A=1; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 2: A=0; B=0; C=1; D=0; E=0; F=1; G=0; break;
  case 3: A=0; B=0; C=0; D=0; E=1; F=1; G=0; break;
  case 4: A=1; B=0; C=0; D=1; E=1; F=0; G=0; break;
  case 5: A=0; B=1; C=0; D=0; E=1; F=0; G=0; break;
  case 6: A=0; B=1; C=0; D=0; E=0; F=0; G=0; break;
  case 7: A=0; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 8: A=0; B=0; C=0; D=0; E=0; F=0; G=0; break;
  case 9: A=0; B=0; C=0; D=0; E=1; F=0; G=0; break;
  case ' ': A=1; B=1; C=1; D=1; E=1; F=1; G=1; break;
  case 'u': A=1; B=1; C=0; D=0; E=0; F=1; G=1; break;
  case 'c': A=1; B=1; C=1; D=0; E=0; F=1; G=0; break;
  case 'b': A=1; B=1; C=0; D=0; E=0; F=0; G=0; break;
  case 't': A=1; B=1; C=1; D=0; E=0; F=0; G=0; break;
  case 'l': A=1; B=1; C=1; D=0; E=0; F=1; G=1; break;
  case 'o': A=1; B=1; C=0; D=0; E=0; F=1; G=0; break;
  case 'A': A=0; B=0; C=0; D=1; E=0; F=0; G=0; break;
  case 'H': A=1; B=0; C=0; D=1; E=0; F=0; G=0; break;
  case 'C': A=0; B=1; C=1; D=0; E=0; F=0; G=1; break;
  case 'O': A=0; B=0; C=0; D=0; E=0; F=0; G=1; break;  
  case 'P': A=0; B=0; C=1; D=1; E=0; F=0; G=0; break;
  case 'E': A=0; B=1; C=1; D=0; E=0; F=0; G=0; break;
  case 'U': A=1; B=0; C=0; D=0; E=0; F=0; G=1; break;
  case 'F': A=0; B=1; C=1; D=1; E=0; F=0; G=0; break;
  case 'L': A=1; B=1; C=1; D=0; E=0; F=0; G=1; break;
  case 'r': A=1; B=1; C=1; D=1; E=0; F=1; G=0; break;
  case '_': A=1; B=1; C=1; D=0; E=1; F=1; G=1; break;
  case 'd': A=1; B=0; C=0; D=0; E=0; F=1; G=0; break;
  case 'n': A=1; B=1; C=0; D=1; E=0; F=1; G=0; break;
  case '': A=0; B=0; C=1; D=1; E=1; F=0; G=1; break;
  default:  A=1; B=1; C=1; D=1; E=1; F=1; G=0; break; //-
 }

   switch (pc)
 {
  case 0: N0=1; if(dpp == 0) {DP=0;} break;
  case 1: N1=1; if(dpp == 1) {DP=0;} break;
  case 2: if(dbuf[2]){N2=1;} if(dpp == 2) {DP=0;} break;
  case 3: if(dbuf[3]){N3=1;} if(dpp == 3) {DP=0;} break;
  default: DP=1; break;
 }
 }
void scan_keys ()
{
#define lp  4
#define lh  10
#define lr  4

static u8 count[6];

if ((!key0)&&(!tkey[0])) {count[0]++;tkey[0]=lr;}
  else if((key0)&&(!tkey[0])) {count[0]=count[1]=0;btn0=rls;}
if ((count[0]>lp)&&(btn0!=hold)) {btn0=prs; count[0]=0; count[1]++;}
//if (count[1]>lh) {btn0=hold; count[1]=0;}

if ((!key1)&&(!tkey[1])) {count[2]++;tkey[1]=lr;}
  else if((key1)&&(!tkey[1])) {count[3]=count[2]=0;btn1=rls;}
if ((count[2]>lp)&&(btn1!=hold)) {btn1=prs; count[2]=0; count[3]++;}
//if (count[3]>lh) {btn1=hold; count[3]=0;}

if ((!key2)&&(!tkey[2])) {count[4]++;tkey[2]=lr;}
  else if((key2)&&(!tkey[2])) {count[5]=count[4]=0;btn2=rls;}
if ((count[4]>0xf0)&&(btn2!=hold)) {btn2=prs; count[4]=0; count[5]++;}
if (count[5]>lh) {btn2=hold; count[5]=0;}
}
struct encoder_struct
{
 s16* value;
 s16  val_min;
 s16  val_max;
 u8   step;
 u8   cycle_form;
}encoder_data; //={&def_tmp,0,9999,1,1}; 
void scan_enc (struct encoder_struct sst)
{
u8 New; bool a1,a0;
static u8 EncState;

if(btn0==rls) a0=1; else a0=0;
if(btn1==rls) a1=1; else a1=0; 
New = (a1<<1)+a0;
 
switch(EncState)
{
case 0:
{
if(New == 2) 
  if(*sst.value<sst.val_max-sst.step)*sst.value+=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_min;
  else *sst.value=sst.val_max;       
if(New == 1) 
  if(*sst.value>sst.val_min+sst.step)*sst.value-=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_max;
  else *sst.value=sst.val_min;
break;
}
case 1:
{
if(New == 0) 
  if(*sst.value<sst.val_max-sst.step)*sst.value+=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_min;
  else *sst.value=sst.val_max;  
if(New == 3) 
  if(*sst.value>sst.val_min+sst.step)*sst.value-=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_max;
  else *sst.value=sst.val_min; 
break;
}
case 2:
{
if(New == 3) 
  if(*sst.value<sst.val_max-sst.step)*sst.value+=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_min;
  else *sst.value=sst.val_max;  
if(New == 0) 
  if(*sst.value>sst.val_min+sst.step)*sst.value-=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_max;
  else *sst.value=sst.val_min;
break;
}
case 3:
{
if(New == 1) 
  if(*sst.value<sst.val_max-sst.step)*sst.value+=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_min;
  else *sst.value=sst.val_max;  
if(New == 2) 
  if(*sst.value>sst.val_min+sst.step)*sst.value-=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_max;
  else *sst.value=sst.val_min;
break;
}
}
 
EncState = New;
}
u16 read_ee_i()
 {
  for(u8 i=0;i<msize;i++)
   {
     if (eeIval[i]!=0xffff) return eeIval[i]; 
   }
  return 7777;
 }
u16 read_ee_u()
 {
  for(u8 i=0;i<msize;i++)
   {
     if (eeUval[i]!=0xffff) return eeUval[i]; 
   }
  return 7777;
 }
void write_ee_i (u16 databyte)
{

 for(u8 i=0;i<msize;i++)
 {
   if(eeIval[i]==databyte) return; 
   
   if(eeIval[i]!=0xffff)
    {
      FLASH_Unlock(FLASH_MEMTYPE_DATA);
      if(i<msize-1)
      {
       eeIval[i+1]=databyte;
       eeIval[i]=0xffff;
      } 
     else 
      {
       eeIval[0]=databyte;
       eeIval[i]=0xffff;
      }
      FLASH_Lock(FLASH_MEMTYPE_DATA); 
      return;
    }
   else if(i==msize-1)
    {
      FLASH_Unlock(FLASH_MEMTYPE_DATA);
      eeIval[0]=databyte;
      FLASH_Lock(FLASH_MEMTYPE_DATA);
    }
 }
 
}
void write_ee_u (u16 databyte)
{

 for(u8 i=0;i<msize;i++)
 {
   if(eeUval[i]==databyte) return; 
   
   if(eeUval[i]!=0xffff)
    {
      FLASH_Unlock(FLASH_MEMTYPE_DATA);
      if(i<msize-1)
      {
       eeUval[i+1]=databyte;
       eeUval[i]=0xffff;
      } 
     else 
      {
       eeUval[0]=databyte;
       eeUval[i]=0xffff;
      }
      FLASH_Lock(FLASH_MEMTYPE_DATA); 
      return;
    }
   else if(i==msize-1)
    {
      FLASH_Unlock(FLASH_MEMTYPE_DATA);
      eeUval[0]=databyte;
      FLASH_Lock(FLASH_MEMTYPE_DATA);
    }
 }
 
}

INTERRUPT_HANDLER(TIM1_X3, 11)
{
  union {u8 adc[2]; u16 adcw;} adcu1,adcu0;
  static u16 Uadc_tmp, Iadc_tmp; 
  
  adcu0.adc[1] = ADC1->DB0RL;
  adcu0.adc[0] = ADC1->DB0RH;
  adcu1.adc[1] = ADC1->DB1RL;
  adcu1.adc[0] = ADC1->DB1RH; //,       

  if(!i_fg)
   {
     adc1val=adcu1.adcw; adc0val=adcu0.adcw;
     Uadc_tmp=Uadc; Iadc_tmp=Iadc;
   }
 
  //PC_ODR_ODR3 = ~PC_ODR_ODR3;

  #define dtv 12 //*62ns
  
  //if(adcu1.adcw>550) {mosfet=0; dev_st=ovl;} //   
  if ((TIM1->CCR1L)&&((adcu0.adcw>Uadc_tmp)||(adcu1.adcw>Iadc_tmp))) {TIM1->CCR1L--;}
  else if ((TIM1->CCR1L<toppwm/2-dtv)&&(adcu0.adcw<Uadc_tmp)&&(adcu1.adcw<Iadc_tmp)){TIM1->CCR1L++;} 
  TIM1->CCR2L = toppwm - TIM1->CCR1L;
  
  ADC1->CSR = 1; ADC1->CR1 |= ADC1_CR1_ADON;
  TIM1->SR1 = 0;
}

INTERRUPT_HANDLER(TIM2_OVF, 13)
{ 
 scan_keys ();
 scan_enc (encoder_data);   
 
 for(u8 i=0;i<timnmbr;i++)
 {
  if(tkey[i])tkey[i]--;
  if(tim[i])tim[i]--;
  if(tmr[i])tmr[i]--;
 }
  IWDG_ReloadCounter();
  TIM2->SR1 = 0; 
}

void devcon(), pwm_on(), pwm_off();

void main()
{
gpio_cfg();
mosfet=0; 
clk_cfg();
adc_cfg();
ITC_SetSoftwarePriority(ITC_IRQ_TIM1_OVF, ITC_PRIORITYLEVEL_3);
ITC_SetSoftwarePriority(ITC_IRQ_TIM2_OVF, ITC_PRIORITYLEVEL_1);
t1_cfg();
t2_cfg();

ADC1->CR1 |= ADC1_CR1_ADON;
if(read_ee_i()==7777)write_ee_i(100);
if(read_ee_u()==7777)write_ee_u(500);

iwdg_cfg();
asm("rim");

  while(1)
  {
    display_data();
    devcon();
  }
}

void assert_failed(uint8_t* file, uint32_t line)
{

}

void devcon()
 {
  static bool itr=0, btn_trg;
   
  if(btn2==rls){btn_trg=0;}
  else if(btn2==hold){btn_trg=1;dev_st=sdn;itr=0;} 

  switch(dev_st) //start,normal,regU,regI,ovl,scp,sdn
  {
   case start:
   {  
    static u8 t;
    if(!itr)
    {
    itr=1; tmr[0]=0xff; t=0x0f;
    dbuf[3]=' ';dbuf[2]='r';dbuf[1]='u';dbuf[0]='n';dpp=0;
    mosfet=1; pwm_on(); Udef=500; Idef=100; 
    }  
   
    if((!tmr[0])&&(t)){tmr[0]=0xff;t--;}
    else if(!t)
     {
       Udef=U_tmp=read_ee_u();
       Idef=I_tmp=read_ee_i();
       i_fg=1;
       Iadc=((u32)Idef*0x1ff)/I_max_adc; 
       Uadc=((u32)Udef*0x3ff)/V_max_adc;
       i_fg=0;
       dev_st=regU;
     }    
    
   if(dev_st!=start){itr=0;}   
   } break;
   
   case regU:
   {  
    if(!itr)
    {
    itr=1; 
    dbuf[3]='U';dpp=1;
    
    encoder_data.val_max=1800;
    encoder_data.val_min=500;
    encoder_data.step=10;
    encoder_data.cycle_form=1;
    encoder_data.value = &def_tmp; 
    i_fg=1;
    Uadc=((u32)Udef*0x3ff)/V_max_adc;  
    i_fg=0;
    } 
    
    if((btn2==prs)&&(!btn_trg)&&(!tim[1])) //      
     {
       btn_trg=1; tim[1]=40000; encoder_data.value = &U_tmp; def_tmp=U_tmp; // ,  
     }
   
    if(tim[1]) //     
     {
       if(def_tmp!=U_tmp){tim[1]=40000;def_tmp=U_tmp;} //         
       if(!tim[0]){tim[0]=1000;convert_data(U_tmp/10);dbuf[3]='u';} //  
       if((btn2==prs)&&(!btn_trg)) //     
        {btn_trg=1;tim[1]=0; dev_st=regI; if(Udef!=U_tmp){write_ee_u(U_tmp);Udef=U_tmp;}} //   
     }
    else //    
     {
       if(!tim[0]) //   
        {
          if(Udef!=U_tmp) //   
           {
             write_ee_u(U_tmp); //   
             Udef=U_tmp;         // 
             i_fg=1;                //    
             Uadc=((u32)Udef*0x3ff)/V_max_adc; //     
             i_fg=0;                //    
             encoder_data.value = &def_tmp; //   
           }
          tim[0]=2500; //    adc
          i_fg=1;
          convert_data(((u32)adc0val*V_max_adc)/(0x3ff*10));dbuf[3]='U'; dpp=1;
          i_fg=0;

        }  
       
     }
    
    
     
    if(!(TIM1->BKR&TIM1_BKR_MOE))dev_st=scp;   //           
    if(dev_st!=regU){itr=0;}   
   } break;

   case regI:
   {  
    if(!itr)
    {
    itr=1; 
    dbuf[3]='C';dpp=1;
    
    encoder_data.val_max=1000;
    encoder_data.val_min=50;
    encoder_data.step=10;
    encoder_data.cycle_form=1;
    encoder_data.value = &def_tmp; 
    i_fg=1;
    Iadc=((u32)Idef*0x1ff)/I_max_adc; 
    i_fg=0;
    } 
    
    if((btn2==prs)&&(!btn_trg)&&(!tim[1]))
     {btn_trg=1; tim[1]=40000;encoder_data.value = &I_tmp; def_tmp=I_tmp;}
   
    if(tim[1])
     {
       if(def_tmp!=I_tmp){tim[1]=40000;def_tmp=I_tmp;}     
       if(!tim[0]){tim[0]=1000;convert_data(I_tmp/10);dbuf[3]='c';} 
       if((btn2==prs)&&(!btn_trg)) 
        {
          btn_trg=1; dev_st=regU; tim[1]=0; 
          if(Idef!=I_tmp){write_ee_i(I_tmp);Idef=I_tmp;}
        } 
     }
    else //    
     {
       if(!tim[0]) //   
        {
          if(Idef!=I_tmp) 
           {
             write_ee_i(I_tmp);
             Idef=I_tmp; 
             i_fg=1;
             Iadc=((u32)Idef*0x1ff)/I_max_adc;
             i_fg=0;
           }
          tim[0]=2500;
          i_fg=1;
          convert_data(((u32)adc1val*I_max_adc)/(500*10));dbuf[3]='C';
          i_fg=0;
        }  
       
     }
    
    
     
    if(!(TIM1->BKR&TIM1_BKR_MOE))dev_st=scp;  
    if(dev_st!=regI){itr=0;}   
   } break;   
   
   case ovl:
   {  
    if(!itr)
    {
     itr=1;
     dbuf[3]=' ';dbuf[2]='O';dbuf[1]='U';dbuf[0]='L';dpp=3;
     mosfet=0; pwm_off();
     tim[1] = 0xffff;
    }  
    
    if((!tim[1])||((btn2==prs)&&(!btn_trg))){btn_trg=1;tim[1]=0;dev_st=start;}
 
    if(dev_st!=ovl){itr=0;}   
   } break;

   case scp:
   {  
    if(!itr)
    {
     itr=1;
     dbuf[3]=' ';dbuf[2]='5';dbuf[1]='C';dbuf[0]='P';dpp=3;
     mosfet=0; pwm_off();
     tim[1] = 0xffff;
    }  
    
    if((!tim[1])||((btn2==prs)&&(!btn_trg))){btn_trg=1;tim[1]=0;dev_st=start;}
   
    if(dev_st!=scp){itr=0;}   
   } break;

   case sdn:
   {  
    if(!itr)
    {
     itr=1;
     dbuf[3]='_';dbuf[2]='_';dbuf[1]='_';dbuf[0]='_';dpp=5;
     mosfet=0; pwm_off();
    }  
    
    if((btn2==prs)&&(!btn_trg)){btn_trg=1;dev_st=start;tim[1]=0;}
    if(dev_st!=sdn){itr=0;}   
   } break;
    
   default:break;   
   }
 }

void pwm_on()
 {
  TIM1->BKR |= TIM1_BKR_MOE; 
 }

void pwm_off()
 {
  TIM1->BKR &= ~TIM1_BKR_MOE; Udef=Idef=0;
 }